from random import choice

from busy.command import QueueCommand


class DropCommand(QueueCommand):
    """Move items to the end of the todo collection of a queue"""

    name = 'drop'
    default_filter = [1]

    def handle_vals(self):
        super().handle_vals()
        if self.selection:
            self.app.ui.send(self.selected_items_list)

    @QueueCommand.wrap
    def execute(self):
        collection = self.app.storage.get_collection(self.queue)
        lolist, hilist = self.collection.split(self.selection)
        self.collection.data = hilist + lolist
        self.status = f"Popped {self.summarize(lolist)}"


class PopCommand(QueueCommand):
    """Move items to the beginning of the collection"""

    name = 'pop'
    default_filter = ['-']

    def handle_vals(self):
        super().handle_vals()
        if len(self.selection) > 1:
            self.app.ui.send(self.selected_items_list)

    @QueueCommand.wrap
    def execute(self):
        collection = self.app.storage.get_collection(self.queue)
        hilist, lolist = self.collection.split(self.selection)
        self.collection.data = hilist + lolist
        self.status = f"Popped {self.summarize(hilist)}"


class PickCommand(QueueCommand):
    """Move a random item to the beginning of the collection"""

    name = 'pick'
    default_filter = ['-']

    @QueueCommand.wrap
    def execute(self):
        index = choice(self.selection)
        item = self.collection[index]
        hilist, lolist = self.collection.split([index])
        self.collection.data = hilist + lolist
        self.status = f"Picked {self.summarize([item])}"



from importlib import import_module
from busy.command import BusyCommand

from wizlib.parser import WizParser

from busy.error import BusyError


class SyncCommand(BusyCommand):
    """Takes the name of an integration"""

    name = 'sync'

    @classmethod
    def add_args(cls, parser: WizParser):
        super().add_args(parser)
        parser.add_argument('integration')

    @BusyCommand.wrap
    def execute(self):
        module = import_module('busy_' + self.integration)
        func = getattr(module, 'sync')
        result = func(self.app)
        self.status = 'Synced'
        return result



from busy.command import CollectionCommand

# Returns the top task from the queue, without tags, a resource, or followons.


class BaseCommand(CollectionCommand):

    name = "base"

    @CollectionCommand.wrap
    def execute(self):
        self.status = f"Listed {self.summarize()}"
        return self.output_items(lambda i: i.base)



from wizlib.parser import WizParser

from busy.command import QueueCommand
from busy.util import date_util


def is_today_or_earlier(plan):
    return plan.plan_date <= date_util.today()


class ActivateCommand(QueueCommand):

    timing: str = ""
    yes: bool = None
    collection_state: str = 'plan'
    name = 'activate'
    default_filter = [is_today_or_earlier]

    @classmethod
    def add_args(cls, parser: WizParser):
        super().add_args(parser)
        # parser.add_argument('--timing', '-t', default='today')
        cls.add_yes_arg(parser)

    def handle_vals(self):
        super().handle_vals()
        if self.selection:
            if not self.provided('yes'):
                items = self.collection.items(self.selection)
                self.app.ui.send(self.selected_items_list)
                intro = f"Activate {self.summarize()}"
                self.confirm(intro)

    @QueueCommand.wrap
    def execute(self):
        todos = self.app.storage.get_collection(self.queue)
        activated = self.collection.delete(self.selection)
        for item in activated:
            item.state = 'todo'
        todos.extend(activated)
        self.status = f"Activated {self.summarize(activated)}"



from wizlib.parser import WizParser
from wizlib.command import CommandCancellation

from busy.command import QueueCommand
from busy.command.defer_command import Deferable
from busy.model.item import Item
from busy.util import date_util
from busy.util.date_util import relative_date


class DoneCommand(Deferable):
    """Combined the old defer and finish commands"""

    name = 'done'

    yes: bool = None
    when: str = ''

    @classmethod
    def add_args(cls, parser: WizParser):
        super().add_args(parser)
        parser.add_argument('--defer', '-d', dest='when')
        cls.add_yes_arg(parser)

    def ask_defer(self):
        self.ask_when()
    ask_defer.name = 'defer'
    ask_defer.key = 'd'

    def ask_other(self):
        self.ask_when()
    ask_other.name = 'other'
    ask_other.key = 'o'

    def handle_vals(self):
        super().handle_vals()
        if self.selection:
            self.app.ui.send(self.selected_items_list)
            self.get_default_when()
            if self.provided('when'):
                if not self.check_when():
                    self.ask_when()
            if not self.provided('yes'):
                if not self.provided('yes'):
                    intro = f"Done {self.summarize()}"
                    if self.plan_date:
                        intro += f" and defer to {self.plan_date}"
                        self.confirm(intro, self.ask_other)
                    else:
                        self.confirm(intro, self.ask_defer)

    @QueueCommand.wrap
    def execute(self):
        date = date_util.today()
        dones = self.app.storage.get_collection(self.queue, 'done')
        plans = self.app.storage.get_collection(self.queue, 'plan')
        items = self.collection.delete(self.selection)
        nexts = [i.done(done_date=date_util.today(),
                        plan_date=self.plan_date) for i in items]
        nexts = [n for n in nexts if n]
        dones += items
        plans += nexts
        if self.plan_date:
            self.status = f"Done for now with {self.summarize(items)}"
            self.status += f" and deferred to {self.plan_date}"
        else:
            self.status = f"Finished {self.summarize(items)}"



from wizlib.parser import WizParser

from busy.command import CollectionCommand
from busy.util.date_util import absolute_date


class ListCommand(CollectionCommand):
    """Show the descriptions with selection numbers, default to all"""

    name = 'list'
    default_filter = ['1-']
    FORMATS = {
        'description': "{!s}",
        'plan_date': "{:%Y-%m-%d}",
        'done_date': "{:%Y-%m-%d}"
    }

    @classmethod
    def add_args(cls, parser: WizParser):
        super().add_args(parser)
        # parser.add_argument('--extended', '-x', action='store_true')
        parser.add_argument('--done_min', '--done-min')
        parser.add_argument('--done_max', '--done-max')

    def handle_vals(self):
        super().handle_vals()
        self.named_filters = []
        if self.provided('done_min'):
            def minfunc(i): return i.done_date >= absolute_date(self.done_min)
            self.named_filters.append(minfunc)
        if self.provided('done_max'):
            def maxfunc(i): return i.done_date <= absolute_date(self.done_max)
            self.named_filters.append(maxfunc)

    @CollectionCommand.wrap
    def execute(self):
        def format(item, index):
            result = f"{(index+1):>6}"
            for colname in self.collection.schema:
                format = self.FORMATS[colname]
                if (colname == 'description'):
                    value = item.listable
                else:
                    value = getattr(item, colname)
                result += f"  {format.format(value)}"
            return result
        self.status = f"Listed {self.summarize()}"
        return self.output_items(format, with_index=True)



from busy.command import CollectionCommand

# Like base but with the tags


class SimpleCommand(CollectionCommand):

    name = "simple"

    @CollectionCommand.wrap
    def execute(self):
        self.status = f"Listed {self.summarize()}"
        return self.output_items(lambda i: i.simple)



# Call the system editor

from io import StringIO
from subprocess import run
from tempfile import NamedTemporaryFile

from busy.model.collection import Collection


def edit_text(text: str, command: str):
    with NamedTemporaryFile(mode="w+") as tempfile:
        tempfile.write(text)
        tempfile.seek(0)
        run([command, tempfile.name])
        tempfile.seek(0)
        return tempfile.read()


def edit_items(collection: Collection, indices, command):
    with StringIO() as oldio:
        collection.write_items(oldio, indices)
        oldio.seek(0)
        oldtext = oldio.read()
    newtext = edit_text(oldtext, command)
    with StringIO() as newio:
        newio.write(newtext)
        newio.seek(0)
        newitems = collection.read_items(newio, indices)
    return newitems



from busy.command import CollectionCommand


class DescribeCommand(CollectionCommand):
    """Show just the description"""

    name = 'describe'

    @CollectionCommand.wrap
    def execute(self):
        self.status = f"Listed {self.summarize()}"
        return self.output_items(lambda i: i.description)



import re
from dataclasses import KW_ONLY, dataclass
from datetime import date, datetime

from busy.util.date_util import absolute_date
from busy.util.date_util import relative_date
from busy.util import date_util


class ItemStateError(Exception):
    pass


START_TIME_FORMAT = '%Y%m%d%H%M'


@dataclass
class Item:

    description: str
    _: KW_ONLY
    state: str = 'todo'
    done_date: date = None
    plan_date: date = None

    def __str__(self):
        """Represent the item as its simple form"""
        return self.simple

    def restricted(*allowed_states):
        """Restrict a method to a specific set of states"""
        def wrapper(method):
            def replacement(self, *args, **kwargs):
                if self.state in allowed_states:
                    return method(self, *args, **kwargs)
                else:
                    raise ItemStateError
            return replacement
        return wrapper

    # Older versions of Busy supported an elaborate "followon" mechanism

    FOLLOW_SPLIT = re.compile(r'\s*\-*\>\s*')
    LEGACY_REPEAT = re.compile(r'^\s*repeat(?:\s+[io]n)?\s+(.+)\s*$', re.I)

    def __setattr__(self, name, value):
        if self.__annotations__[name] == date:
            value = absolute_date(value)
        super().__setattr__(name, value)

    @property
    def _words(self):
        return self.body.split()

    @property
    def tags(self):
        wins = [w for w in self._words if w.startswith("#")]
        return {w[1:].lower() for w in wins}

    @property
    def resource(self):
        wins = [w for w in self._words if w.startswith("@")]
        return wins.pop()[1:] if wins else ""

    @property
    def data(self):
        return {w[1]: w[2:] for w in self._words if w.startswith("!")}

    @property
    def start_time(self):
        if 's' in self.data:
            return datetime.strptime(self.data['s'], START_TIME_FORMAT)

    @property
    def elapsed_time(self):
        if 'e' in self.data:
            return int(self.data['e'])
        else:
            return 0

    @property
    def base(self):
        """body description with no tags, resource, or data"""
        wins = [w for w in self._words if w[0] not in '#@!']
        return " ".join(wins)

    @property
    def simple(self):
        """Base plus tags"""
        wins = [w for w in self._words if w[0] not in '@!']
        return " ".join(wins)

    @property
    def listable(self):
        """Simple plus repeat"""
        if self.repeat_text:
            return f"{self.simple} > {self.repeat_text}"
        else:
            return self.simple

    @property
    def nodata(self):
        """Everything but the bang data"""
        wins = [w for w in self._words if w[0] not in '!']
        return " ".join(wins)

    @property
    def repeat_text(self):
        """Second and successive segments, all but body"""
        split = self.FOLLOW_SPLIT.split(self.description, maxsplit=1)
        if len(split) > 1:
            next = split[1]
            # Legacy "repeat" text
            match = self.LEGACY_REPEAT.match(next)
            if match:
                return match.group(1)
            else:
                return next
        else:
            return ""

    @property
    def body(self):
        """Without the repeat followon"""
        split = self.FOLLOW_SPLIT.split(self.description, maxsplit=1)
        if split:
            return split[0]
        else:
            return ""

    @property
    def repeat_date(self):
        text = self.repeat_text
        if text:
            return relative_date(text)

    @restricted('todo')
    def done(self, done_date: date, plan_date: date = None):
        """Updates the item to done and returns a copy as a plan for the
        plan_date if provided"""
        self.state = 'done'
        self.done_date = done_date
        plan_description = self.nodata
        if self.repeat_text:
            plan_description += f" > {self.repeat_text}"
        self.description = self.body
        if plan_date:
            return Item(plan_description, state='plan', plan_date=plan_date)

    @restricted('done')
    def undone(self):
        self.state = 'todo'

    @restricted('todo')
    def plan(self, plan_date: date):
        self.state = 'plan'
        self.plan_date = plan_date

    @restricted('plan')
    def unplan(self):
        self.state = 'todo'

    @restricted('todo')
    def update_time(self):
        """Update elapsed time based on start time"""
        if self.start_time and self.base:
            prev = self.elapsed_time
            new = (date_util.now() - self.start_time).seconds // 60
            elapsed = new + prev
            changed = f"{self.nodata} !e{elapsed}"
            if self.repeat_text:
                changed += f" > {self.repeat_text}"
            self.description = changed

    @restricted('todo')
    def start_timer(self):
        if self.base and not self.start_time:
            start = date_util.now().strftime(START_TIME_FORMAT)
            elapsed = self.elapsed_time
            changed = f"{self.nodata} !s{start}"
            if elapsed:
                changed += f" !e{elapsed}"
            if self.repeat_text:
                changed += f" > {self.repeat_text}"
            self.description = changed



# Treat a collection as an in-memory database of Items, and this is the query
# engine.
#
# "Groups" create some degree of logic -- all the filters within a group are
# OR'd with each other, then groups are AND'ed against other groups.

import re


class Selector:

    CRITERIUM_TYPES = []

    @classmethod
    def add_criterium_type(self, criterium_type):
        self.CRITERIUM_TYPES.append(criterium_type)

    @classmethod
    def get_criterium(self, word):
        for criterium_type in self.CRITERIUM_TYPES:
            if criterium_type.match(word):
                return criterium_type(word)
        raise RuntimeError(f"Invalid selection criterium {word}")

    def __init__(self, args):
        """Receives filters as arguments and initializes the selector"""
        result = {}
        for arg in args:
            criterium = self.get_criterium(arg)
            if hasattr(criterium, 'group'):
                groupnum = criterium.group
            else:
                # 9 arbitrarily high
                groupnum = max(max(result.keys() | [0]), 9) + 1
            if groupnum not in result:
                result[groupnum] = []
            result[groupnum].append(criterium)
        self.filter_groups = result

    def hit(self, index, value, count):
        """Return true if the value at the index matches the criteria logic"""
        for group in self.filter_groups:
            filters = self.filter_groups[group]
            hit = any(f.hit(index, value, count) for f in filters)
            if not hit:
                return False
        return True

    def indices(self, elements):
        """Return the indices of the elements that match the criteria"""
        enumeration = enumerate(elements)
        count = len(elements)
        return [i for i, t in enumeration if self.hit(i, t, count)]


class NumberCriterium:

    group = 1

    def match(word):
        return str(word).isdigit()

    def __init__(self, word):
        self.index = int(word) - 1

    def hit(self, index, value, count):
        return index == self.index


Selector.add_criterium_type(NumberCriterium)


class RangeCriterium:

    group = 1
    MATCH = r'^(\d+\-\d*)|\-$'

    @classmethod
    def match(cls, word):
        return isinstance(word, str) and bool(re.match(cls.MATCH, word))

    def __init__(self, word):
        split = word.split('-')
        self.start = int(split[0]) - 1 if split[0] else None
        self.end = int(split[1]) - 1 if split[1] else None

    def hit(self, index, value, count):
        if self.start is None and self.end is None:
            return index == count - 1
        elif self.end is None:
            return index >= self.start
        else:
            return index >= self.start and index <= self.end


Selector.add_criterium_type(RangeCriterium)


class TagCriterium:

    group = 2

    def match(word):
        return isinstance(
            word, str) and bool(
            re.match(r'^[a-zA-Z\-]+$', word))

    def __init__(self, word):
        self.tag = str(word).lower()

    def hit(self, index, value, count):
        return ('#' + self.tag) in str(value).lower().split()


Selector.add_criterium_type(TagCriterium)


class CombinedTagCriterium:

    group = 2

    MATCH = r'^[a-zA-Z\-]+(\+[a-zA-Z\-]+)+$'
    FIND = r'([a-zA-Z\-]+)'

    @classmethod
    def match(cls, word):
        if isinstance(word, str):
            return bool(re.match(cls.MATCH, word))

    def __init__(self, word):
        tags = re.findall(self.FIND, word)
        self.tags = [t.lower() for t in tags]

    def hit(self, index, value, count):
        words = str(value).lower().split()
        return all(('#'+t in words) for t in self.tags)


Selector.add_criterium_type(CombinedTagCriterium)


class FunctionCriterium:
    """Evaluates a function against the value (not the index) - used for named
    filters - each gets its own group"""

    def match(arg):
        return callable(arg)

    def __init__(self, arg):
        self.func = arg

    def hit(self, index, value, count):
        return self.func(value)


Selector.add_criterium_type(FunctionCriterium)



import sys
from wizlib.app import WizApp
from wizlib.ui_handler import UIHandler
from wizlib.config_handler import ConfigHandler

from busy.command import BusyCommand
from busy.storage.file_storage import FileStorage


class BusyApp(WizApp):

    base = BusyCommand
    name = 'busy'
    handlers = [UIHandler, ConfigHandler]

    def __init__(self, **handlers):
        super().__init__(**handlers)
        self.storage = FileStorage(self.config.get('busy-storage-directory'))

    def run(self, **vals):
        super().run(**vals)
        self.output_current_task()

    def output_current_task(self):
        """Just output the current task after a command"""
        collection = self.storage.get_collection(
            'tasks', 'todo')
        if len(collection):
            item = collection[0].simple
            self.ui.send(item)
